"""Loader that uses unstructured to load files."""
import collections
from abc import ABC, abstractmethod
from typing import IO, Any, Dict, List, Sequence, Union

from .base import BaseLoader
from ..docstore.document import Document


def satisfies_min_unstructured_version(min_version: str) -> bool:
    """Checks to see if the installed unstructured version exceeds the minimum version
    for the feature in question."""
    from unstructured.__version__ import __version__ as __unstructured_version__

    min_version_tuple = tuple([int(x) for x in min_version.split(".")])

    # NOTE(MthwRobinson) - enables the loader to work when you're using pre-release
    # versions of unstructured like 0.4.17-dev1
    _unstructured_version = __unstructured_version__.split("-")[0]
    unstructured_version_tuple = tuple(
        [int(x) for x in _unstructured_version.split(".")]
    )

    return unstructured_version_tuple >= min_version_tuple


def validate_unstructured_version(min_unstructured_version: str) -> None:
    """Raises an error if the unstructured version does not exceed the
    specified minimum."""
    if not satisfies_min_unstructured_version(min_unstructured_version):
        raise ValueError(
            f"unstructured>={min_unstructured_version} is required in this loader."
        )


class UnstructuredBaseLoader(BaseLoader, ABC):
    """Loader that uses unstructured to load files."""

    def __init__(self, mode: str = "single", **unstructured_kwargs: Any):
        """Initialize with file path."""
        try:
            import unstructured  # noqa:F401
        except ImportError:
            raise ValueError(
                "unstructured package not found, please install it with "
                "`pip install unstructured`"
            )
        _valid_modes = {"single", "elements", "paged"}
        if mode not in _valid_modes:
            raise ValueError(
                f"Got {mode} for `mode`, but should be one of `{_valid_modes}`"
            )
        self.mode = mode

        if not satisfies_min_unstructured_version("0.5.4"):
            if "strategy" in unstructured_kwargs:
                unstructured_kwargs.pop("strategy")

        self.unstructured_kwargs = unstructured_kwargs

    @abstractmethod
    def _get_elements(self) -> List:
        """Get elements."""

    @abstractmethod
    def _get_metadata(self) -> dict:
        """Get metadata."""

    def load(self) -> List[Document]:
        """Load file."""
        elements = self._get_elements()
        if self.mode == "elements":
            docs: List[Document] = list()
            for element in elements:
                metadata = self._get_metadata()
                # NOTE(MthwRobinson) - the attribute check is for backward compatibility
                # with unstructured<0.4.9. The metadata attributed was added in 0.4.9.
                if hasattr(element, "metadata"):
                    metadata.update(element.metadata.to_dict())
                if hasattr(element, "category"):
                    metadata["category"] = element.category
                docs.append(Document(page_content=str(element), metadata=metadata))
        elif self.mode == "paged":
            text_dict: Dict[int, str] = {}
            meta_dict: Dict[int, Dict] = {}

            for idx, element in enumerate(elements):
                metadata = self._get_metadata()
                if hasattr(element, "metadata"):
                    metadata.update(element.metadata.to_dict())
                page_number = metadata.get("page_number", 1)

                # Check if this page_number already exists in docs_dict
                if page_number not in text_dict:
                    # If not, create new entry with initial text and metadata
                    text_dict[page_number] = str(element) + "\n\n"
                    meta_dict[page_number] = metadata
                else:
                    # If exists, append to text and update the metadata
                    text_dict[page_number] += str(element) + "\n\n"
                    meta_dict[page_number].update(metadata)

            # Convert the dict to a list of Document objects
            docs = [
                Document(page_content=text_dict[key], metadata=meta_dict[key])
                for key in text_dict.keys()
            ]
        elif self.mode == "single":
            metadata = self._get_metadata()
            text = "\n\n".join([str(el) for el in elements])
            docs = [Document(page_content=text, metadata=metadata)]
        else:
            raise ValueError(f"mode of {self.mode} not supported.")
        return docs


class UnstructuredFileLoader(UnstructuredBaseLoader):
    """Loader that uses unstructured to load files."""

    def __init__(
        self,
        file_path: Union[str, List[str]],
        mode: str = "single",
        **unstructured_kwargs: Any,
    ):
        """Initialize with file path."""
        self.file_path = file_path
        super().__init__(mode=mode, **unstructured_kwargs)

    def _get_elements(self) -> List:
        from unstructured.partition.auto import partition

        return partition(filename=self.file_path, **self.unstructured_kwargs)

    def _get_metadata(self) -> dict:
        return {"source": self.file_path}


def get_elements_from_api(
    file_path: Union[str, List[str], None] = None,
    file: Union[IO, Sequence[IO], None] = None,
    api_url: str = "https://api.unstructured.io/general/v0/general",
    api_key: str = "",
    **unstructured_kwargs: Any,
) -> List:
    """Retrieves a list of elements from the Unstructured API."""
    if isinstance(file, collections.abc.Sequence) or isinstance(file_path, list):
        from unstructured.partition.api import partition_multiple_via_api

        _doc_elements = partition_multiple_via_api(
            filenames=file_path,
            files=file,
            api_key=api_key,
            api_url=api_url,
            **unstructured_kwargs,
        )

        elements = []
        for _elements in _doc_elements:
            elements.extend(_elements)

        return elements
    else:
        from unstructured.partition.api import partition_via_api

        return partition_via_api(
            filename=file_path,
            file=file,
            api_key=api_key,
            api_url=api_url,
            **unstructured_kwargs,
        )


class UnstructuredAPIFileLoader(UnstructuredFileLoader):
    """Loader that uses the unstructured web API to load files."""

    def __init__(
        self,
        file_path: Union[str, List[str]] = "",
        mode: str = "single",
        url: str = "https://api.unstructured.io/general/v0/general",
        api_key: str = "",
        **unstructured_kwargs: Any,
    ):
        """Initialize with file path."""

        if isinstance(file_path, str):
            validate_unstructured_version(min_unstructured_version="0.6.2")
        else:
            validate_unstructured_version(min_unstructured_version="0.6.3")

        self.url = url
        self.api_key = api_key

        super().__init__(file_path=file_path, mode=mode, **unstructured_kwargs)

    def _get_metadata(self) -> dict:
        return {"source": self.file_path}

    def _get_elements(self) -> List:
        return get_elements_from_api(
            file_path=self.file_path,
            api_key=self.api_key,
            api_url=self.url,
            **self.unstructured_kwargs,
        )


class UnstructuredFileIOLoader(UnstructuredBaseLoader):
    """Loader that uses unstructured to load file IO objects."""

    def __init__(
        self,
        file: Union[IO, Sequence[IO]],
        mode: str = "single",
        **unstructured_kwargs: Any,
    ):
        """Initialize with file path."""
        self.file = file
        super().__init__(mode=mode, **unstructured_kwargs)

    def _get_elements(self) -> List:
        from unstructured.partition.auto import partition

        return partition(file=self.file, **self.unstructured_kwargs)

    def _get_metadata(self) -> dict:
        return {}


class UnstructuredAPIFileIOLoader(UnstructuredFileIOLoader):
    """Loader that uses the unstructured web API to load file IO objects."""

    def __init__(
        self,
        file: Union[IO, Sequence[IO]],
        mode: str = "single",
        url: str = "https://api.unstructured.io/general/v0/general",
        api_key: str = "",
        **unstructured_kwargs: Any,
    ):
        """Initialize with file path."""

        if isinstance(file, collections.abc.Sequence):
            validate_unstructured_version(min_unstructured_version="0.6.3")
        if file:
            validate_unstructured_version(min_unstructured_version="0.6.2")

        self.url = url
        self.api_key = api_key

        super().__init__(file=file, mode=mode, **unstructured_kwargs)

    def _get_elements(self) -> List:
        return get_elements_from_api(
            file=self.file,
            api_key=self.api_key,
            api_url=self.url,
            **self.unstructured_kwargs,
        )



import logging
import random
from typing import Iterable

from agere.commander import PASS_WORD, Job, Callback, tasker

from .blinker_wrapper import async_wrapper_with_loop
from .context import Context
from .role import Role
from .signals import notification_signal, common_message_signal
from ..gptui_kernel.manager import ManagerInterface


gptui_logger = logging.getLogger("gptui_logger")


class ResponseJob(Job):
    def __init__(
        self,
        manager: ManagerInterface,
        response: Iterable,
        context: Context,
        callback: Callback | None = None,
        at_receiving_start: list[dict] | None = None,
        at_receiving_end: list[dict] | None = None,
    ):
        super().__init__(callback=callback)
        self.response = response
        self.context = context
        self.manager = manager
        self.at_receiving_start = at_receiving_start
        self.at_receiving_end = at_receiving_end

    @tasker(PASS_WORD)
    async def task(self):
        ResponseHandler = self.manager.get_handler("ResponseHandler")
        handler = ResponseHandler(self.manager, self.context).handle_response(
            response=self.response,
            at_receiving_start=self.at_receiving_start,
            at_receiving_end=self.at_receiving_end,
        )
        return handler


class GroupTalkManager(Job):
    def __init__(self, manager: ManagerInterface, user_name: str = "admin"):
        self._speaking = None
        self.running = False
        self.manager = manager
        self.user_name = user_name
        self.group_talk_manager_id: int | None = None
        self.roles: dict[str, Role] = {}
        self.user_talk_buffer = []
        super().__init__()
    
    @property
    def speaking(self) -> str | None:
        return self._speaking

    @speaking.setter
    def speaking(self, value: str | None):
        if value is None:
            messages = [{"role": "user", "name": self.user_name, "content": message} for message in self.user_talk_buffer]
            if messages:
                common_message_signal.send(
                    self,
                    _async_wrapper=async_wrapper_with_loop,
                    message={
                        "content": messages,
                        "flag": "group_talk_user_message_send",
                    },
                )
                response_dict = {}
                items = list(self.roles.items())
                # Randomly shuffle the order to give each role an equal opportunity to speak.
                random.shuffle(items)
                for role_name, role in items:
                    response_dict[role_name] = role.chat(message=messages)
                GroupTalkHandler = self.manager.get_handler("GroupTalkHandler")
                # Because at any given moment only one role can speak, it is safe when multiple dialogue tasks are running in parallel;
                # there will be no inconsistency in the role's chat history.
                self.call_handler(GroupTalkHandler().handle_response(response_dict))
                self.user_talk_buffer = []
        self._speaking = value

    def create_role(self, role: Role, role_name: str) -> bool:
        if role_name in self.roles:
            notification_signal.send(
                self,
                _async_wrapper=async_wrapper_with_loop,
                message={
                    "content": f"The name of '{role_name}' has already existed.",
                    "flag": "warning",
                }
            )
            return False
        self.roles[role_name] = role
        return True

    def close_group_talk(self):
        self.running = False
        
    @tasker(PASS_WORD)
    async def task(self):
        self.running = True
        GroupTalkHandler = self.manager.get_handler("GroupTalkHandler")
        return GroupTalkHandler().wait_for_termination(self)


class TalkToAll(Job):
    def __init__(self, message_content: str, message_from: str):
        self.message_content = message_content
        self.message_from = message_from
        super().__init__()

    @tasker(PASS_WORD)
    async def task(self):
        talk_manager = self.ancestor_chain[-2]
        assert isinstance(talk_manager, GroupTalkManager)
        if not talk_manager.running:
            return
        response_dict = {}
        all_roles = list(talk_manager.roles.items())
        # Randomly shuffle the order to give each role an equal opportunity to speak.
        random.shuffle(all_roles)
        try:
            for role_name, role in all_roles:
                if role_name == self.message_from:
                    response_dict[role_name] = role.chat(message={"role": "assistant", "content": self.message_content})
                else:
                    response_dict[role_name] = role.chat(message={"role": "user", "name": self.message_from, "content": self.message_content})
        except Exception as e:
            gptui_logger.info(f"Encountered an error when the group chat member {self.message_from} talks to all, error: {e}")
        finally:
            talk_manager.speaking = None
        GroupTalkHandler = talk_manager.manager.get_handler("GroupTalkHandler")
        return GroupTalkHandler().handle_response(response_dict)



import logging
from typing import Iterable

from openai.types.chat import ChatCompletionMessageParam

from .context import BeadOpenaiContext, OpenaiContext
from .openai_error import OpenaiErrorHandler
from .openai_tokens_truncate import trim_excess_tokens
from .utils.openai_api import openai_api_client
from ..gptui_kernel.manager import ManagerInterface


gptui_logger = logging.getLogger("gptui_logger")


class Role:
    def __init__(self, name: str, manager: ManagerInterface, openai_context_parent: OpenaiContext):
        """Role use the same openai parameters as in the parent conversation.
        """
        self.name = name
        self.context = BeadOpenaiContext(parameters=openai_context_parent.parameters)
        self.manager = manager
        self.openai_api_client = openai_api_client(manager.dot_env_config_path)
        self.context.max_sending_tokens_num = openai_context_parent.max_sending_tokens_num
        self.openai_context_parent = openai_context_parent
        self.context.chat_context_saver = "inner"

    def set_role_prompt(self, prompt: str):
        self.context.bead = [{"role": "system", "content": prompt}]
        self.context.insert_bead()

    def chat(self, message: ChatCompletionMessageParam | list[ChatCompletionMessageParam]) -> Iterable:
        self.context.auto_insert_bead()
        if isinstance(message, dict):
            self.context.chat_context_append(message=message)
        else:
            for one_message in message:
                self.context.chat_context_append(message=one_message)
        self.context.parameters = self.openai_context_parent.parameters.copy()
        self.context.parameters["stream"] = True
        trim_messages = trim_excess_tokens(self.context, offset=0)
        try:
            response = self.openai_api_client.with_options(timeout=20.0).chat.completions.create(
                messages=trim_messages,
                **self.context.parameters,
            )
        except Exception as e:
            OpenaiErrorHandler().openai_error_handle(error=e, context=self.context, event_loop=True)
            raise e
        return response



import logging
import os

from textual import events
from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen, Screen
from textual.widgets import Button, Label, Input, DirectoryTree, MarkdownViewer

from ..utils.my_text import MyText as Text


gptui_logger = logging.getLogger("gptui_logger")


class CheckDialog(ModalScreen[bool]):
    """Check an action."""

    CSS = """
    CheckDialog {
        align: center middle;
    }
    CheckDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 60;
        height: 11;
        border: thick $background 80%;
        background: $surface;
    }
    CheckDialog #prompt {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }
    CheckDialog Button{
        width: 100%;
    }
    """

    def __init__(self, prompt: str | Text = "Are you sure?", yes_label: str = "Yse", no_label: str = "No", *args, **kwargs):
        self.prompt = prompt
        self.yes_label = yes_label
        self.no_label = no_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(self.prompt, id="prompt"),
            Button(self.yes_label, id="yes"),
            Button(self.no_label, id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            self.dismiss(True)
        elif event.button.id == "no":
            self.dismiss(False)


class InputDialog(ModalScreen[tuple[bool, str]]):
    """Ask whether to close the conversation."""

    CSS = """
    InputDialog {
        align: center middle;
    }
    InputDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 2fr 3;
        padding: 0 1;
        width: 60;
        height: 11;
        border: thick $background 80%;
        background: $surface;
    }
    InputDialog #input {
        column-span: 2;
        height: 1fr;
        width: 1fr;
    }
    InputDialog #prompt {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }
    InputDialog Button{
        width: 100%;
    }
    """

    def __init__(self, prompt: str | Text = "Input:", placeholder: str = "", default_input: str | None = None, ok_label: str = "OK", cancel_label: str = "Cancel", *args, **kwargs):
        self.prompt = prompt
        self.placeholder = placeholder
        self.default_input = default_input
        self.ok_label = ok_label
        self.cancel_label = cancel_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self.input = Input(placeholder=self.placeholder, id="input")
        yield Grid(
            Label(self.prompt, id="prompt"),
            self.input,
            Button(self.ok_label, id="ok"),
            Button(self.cancel_label, id="cancel"),
            id="dialog",
        )
    
    def on_mount(self):
        if self.default_input:
            self.input.value = self.default_input

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "ok":
            self.dismiss((True, self.input.value))
        elif event.button.id == "cancel":
            self.dismiss((False, self.input.value))


class SelectPathDialog(ModalScreen[tuple[bool, str]]):
    """Ask whether to close the conversation."""

    CSS = """
    SelectPathDialog {
        align: center middle;
    }
    SelectPathDialog > #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1 1fr 2fr 3;
        padding: 0 1;
        width: 60;
        height: 20;
        border: thick $background 80%;
        background: $surface;
    }
    SelectPathDialog #input {
        column-span: 2;
        height: 1;
        width: 1fr;
    }
    SelectPathDialog #directory_tree {
        column-span: 2;
        height: 1fr;
        width: 1fr;
    }
    SelectPathDialog #prompt {
        column-span: 2;
        height: 1;
        width: 1fr;
        content-align: center middle;
    }
    SelectPathDialog Button{
        width: 100%;
    }
    """

    def __init__(
        self,
        prompt: str | Text = "Input:",
        placeholder: str = "",
        ok_label: str = "OK",
        cancel_label: str = "Cancel",
        *args,
        **kwargs
        ):
        self.prompt = prompt
        self.placeholder = placeholder
        self.ok_label = ok_label
        self.cancel_label = cancel_label
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        self.input = Input(placeholder=self.placeholder, id="input")
        yield Grid(
            Label(self.prompt, id="prompt"),
            self.input,
            DirectoryTree("./user", id="directory_tree"),
            Button(self.ok_label, id="ok"),
            Button(self.cancel_label, id="cancel"),
            id="dialog",
        )

    def on_tree_node_highlighted(self, event) -> None:
        selected_path = event.node.data.path
        if selected_path.is_file():
            self.input.value = str(selected_path.parent) + os.sep
        if selected_path.is_dir():
            self.input.value = str(selected_path) + os.sep
    
    async def on_input_submitted(self, message) -> None:
        self.dismiss((True, message.value))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "ok":
            self.dismiss((True, self.input.value))
        elif event.button.id == "cancel":
            self.dismiss((False, self.input.value))


class MarkdownPreview(ModalScreen):
    
    CSS = """
    MarkdownPreview {
        align: center middle;
    }
    MarkdownPreview MarkdownViewer {
        width: 50%;
        height: 50%;
        border: ascii $accent;
    }
    """

    BINDINGS = [
        ("space", "quit", "Quit"),
    ]

    def __init__(self, markdown: str | None = None, previewer_title: str | None = None, *args, **kwargs):
        self.markdown = markdown
        self.previewer_title = previewer_title
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        previewer = self.query_one(MarkdownViewer)
        if self.previewer_title is not None:
            previewer.border_title = self.previewer_title
        previewer.border_subtitle = "Press 'space' to close this previewer"
    
    def compose(self) -> ComposeResult:
        yield MarkdownViewer(self.markdown, show_table_of_contents=False)

    def action_quit(self) -> None:
        self.app.pop_screen()


class HotKey(ModalScreen[str]):

    CSS = """
    HotKey {
        align: center bottom;
    }
    HotKey Label {
        width: 80;
        height: 12;
        text-align: center;
        border: ascii $accent;
    }
    """

    BINDINGS = [
        ("escape,ctrl+underscore", "quit", "Quit"),
    ]

    def __init__(self, display_content: str | Text, *args, **kwargs):
        self.display_content = display_content
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        yield Label(self.display_content)

    def action_quit(self) -> None:
        self.app.pop_screen()

    def on_key(self, event: events.Key) -> None:
        self.dismiss(event.key)
